/*
 * generated by Xtext
 */
package com.framsticks.framclipse.scoping;

import static com.framsticks.framclipse.framScript.FramScriptPackage.Literals.ARRAY_ELEMENT_EXPRESSION__ARRAY;
import static com.framsticks.framclipse.framScript.FramScriptPackage.Literals.QUALIFIED_EXPRESSION__CHILD;
import static com.framsticks.framclipse.framScript.FramScriptPackage.Literals.QUALIFIED_EXPRESSION__PARENT;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.impl.ResourceImpl;
import org.eclipse.xtext.resource.EObjectDescription;
import org.eclipse.xtext.resource.IEObjectDescription;
import org.eclipse.xtext.scoping.IScope;
import org.eclipse.xtext.scoping.Scopes;
import org.eclipse.xtext.scoping.impl.AbstractDeclarativeScopeProvider;
import org.eclipse.xtext.scoping.impl.SimpleScope;

import com.framsticks.framclipse.framScript.Block;
import com.framsticks.framclipse.framScript.Code;
import com.framsticks.framclipse.framScript.FramScriptFactory;
import com.framsticks.framclipse.framScript.Function;
import com.framsticks.framclipse.framScript.GotoStatment;
import com.framsticks.framclipse.framScript.Invocation;
import com.framsticks.framclipse.framScript.LabeledStatement;
import com.framsticks.framclipse.framScript.Model;
import com.framsticks.framclipse.framScript.ProposableFunctionImpl;
import com.framsticks.framclipse.framScript.ProposableVariableDeclarationImpl;
import com.framsticks.framclipse.framScript.QualifiedExpression;
import com.framsticks.framclipse.framScript.Statement;
import com.framsticks.framclipse.framScript.VariableDeclaration;
import com.framsticks.framclipse.framScript.VariableDeclarations;
import com.framsticks.framclipse.framScript.VariableRef;
import com.framsticks.framclipse.script.ConstantProvider;
import com.framsticks.framclipse.script.ExpressionTraverser;
import com.framsticks.framclipse.script.model.Element;
import com.framsticks.framclipse.script.model.Type;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.inject.Inject;

/**
 * This class contains custom scoping description.
 * 
 * see : http://www.eclipse.org/Xtext/documentation/latest/xtext.html#scoping
 * on how and when to use it 
 *
 */
public class FramScriptScopeProvider extends AbstractDeclarativeScopeProvider {
	
	private final ConstantProvider provider;
	
	private final ExpressionTraverser traverser;
	
	private Resource resource;
	
	private Map<String, IScope> typesScopes = Maps.newHashMap();
	private Map<Type, IScope> functionScopes = Maps.newHashMap();
	private Map<Type, IScope> fieldScopes = Maps.newHashMap();
	
	@Inject
	public FramScriptScopeProvider(ConstantProvider provider, ExpressionTraverser traverser) {
		this.provider = provider;
		this.traverser = traverser;
		this.resource = new ResourceImpl(URI.createURI("/"));
		initTypes();
	}
	
	protected void initTypes() {
		List<Function> allFunctions = Lists.newArrayList();
		List<VariableDeclaration> allFields = Lists.newArrayList();
		
		for (Entry<String, Map<String, Type>> entry : provider.getTypes().entrySet()) {
			List<VariableDeclaration> types = Lists.newArrayList();	
			for (Type type : entry.getValue().values()) {
				types.add(createType(type));
				
				ArrayList<Function> functions = Lists.newArrayList();
				ArrayList<VariableDeclaration> fields = Lists.newArrayList();
				for (Element element : type.getElements()) {
					if (element.isFunction()) {
						functions.add(createFunction(element));
					} else {
						fields.add(createField(element));
					}
				}
				functionScopes.put(type, Scopes.scopeFor(functions));
				fieldScopes.put(type, Scopes.scopeFor(fields));
				allFunctions.addAll(functions);
				allFields.addAll(fields);
			}
			typesScopes.put(entry.getKey(), Scopes.scopeFor(types));
		}
		
		functionScopes.put(null, Scopes.scopeFor(allFunctions));
		fieldScopes.put(null, Scopes.scopeFor(allFields));
	}
	
	protected IScope scope_Invocation_function(Invocation inv, EReference ref) {
		QualifiedExpression parent = getParent(inv);
		IScope scope;
		if (parent != null) {
			scope = scope_Invocation_function(parent, ref);
		} else {
			scope = scope_Invocation_function(getFunction(inv), ref);
		}
		scope = filterByArgumentsCount(scope, inv.getArgs().size());
		return scope;
	}

	protected IScope scope_Invocation_function(QualifiedExpression inv, EReference ref) {
		Type type = traverser.getLastType(inv);
		IScope scope = functionScopes.get(type);
		return scope != null ? scope : IScope.NULLSCOPE;
	}

	protected IScope scope_Invocation_function(Function f, EReference ref) {
		IScope scope = delegateGetScope(f, ref);
		scope = addAliases(scope);
		return scope;
	}

	private IScope filterByArgumentsCount(IScope scope, int size) {
		List<IEObjectDescription> filtered = Lists.newArrayList();
		for (IEObjectDescription desc : scope.getAllElements()) {
			Function function = (Function) desc.getEObjectOrProxy();
			if (function.getParams().size() == size) {
				filtered.add(desc);
			}
		}
		return new SimpleScope(filtered);
	}
	
	private IScope addAliases(IScope scope) {
		Iterable<IEObjectDescription> contents = scope.getAllElements();
		List<IEObjectDescription> withAliases = Lists.newArrayList(contents);
		for (IEObjectDescription desc : contents) {
			Function function = (Function) desc.getEObjectOrProxy();
			for (String string : function.getAliases()) {
				withAliases.add(EObjectDescription.create(string, function));
			}
		}
		return new SimpleScope(withAliases);
	}

	protected IScope scope_VariableRef_var(VariableRef var, EReference ref) {
		QualifiedExpression parent = getParent(var);
		if (parent != null) {
			return scope_VariableRef_var(parent, ref);
		} else {
			return scope_VariableRef_var(getBlock(var), ref);
		}
	}

	protected IScope scope_VariableRef_var(QualifiedExpression var, EReference ref) {
		Type type = traverser.getLastType(var);
		IScope scope = fieldScopes.get(type);
		return scope != null ? scope : IScope.NULLSCOPE;
	}
	
	protected IScope scope_VariableRef_var(Block block, EReference ref) {
		List<VariableDeclaration> vars = Lists.newArrayList();
		for (Statement stmt : block.getStatements()) {
			if (stmt instanceof VariableDeclarations) {
				vars.addAll(((VariableDeclarations) stmt).getVars());
			}
		}
		return Scopes.scopeFor(vars, getScope(block.eContainer(), ref));
	}
	
	protected IScope scope_VariableRef_var(Function f, EReference ref) {
		List<VariableDeclaration> vars = Lists.newArrayList();
		vars.addAll(f.getParams());
		return Scopes.scopeFor(vars, getScope(f.eContainer(), ref));
	}
	
	protected IScope scope_VariableRef_var(Code code, EReference ref) {
		List<VariableDeclaration> vars = Lists.newArrayList();
		for (VariableDeclarations declarations : code.getGlobals()) {
			vars.addAll(declarations.getVars());
		}
		return Scopes.scopeFor(vars, getScope(code.eContainer(), ref));
	}
	
	protected IScope scope_VariableRef_var(Model model, EReference ref) {
//		IScope outerScope = delegateGetScope(model, ref).getOuterScope();
		IScope types = typesScopes.get(getExtension(model));
		return new SimpleScope(types.getAllElements());
	}

	protected IScope scope_GotoStatment_dest(GotoStatment stmt, EReference ref) {
		Block code = getFunction(stmt).getCode();
		List<LabeledStatement> labels = Lists.newArrayList();
		for (TreeIterator<EObject> i = code.eAllContents(); i.hasNext();) {
			EObject object = i.next();
			if (object instanceof LabeledStatement) {
				labels.add((LabeledStatement) object);
			}
		}
		return Scopes.scopeFor(labels);
	}
	
	private Function createFunction(Element element) {
		Function function = new ProposableFunctionImpl(element);
		EList<VariableDeclaration> params = function.getParams();
		for (int i = 0; i < element.getArguments().size(); i++) {
			params.add(createVariableDeclaration(element.getArguments().get(i).getName(i)));
		}
		resource.getContents().add(function);
		return function;
	}

	private VariableDeclaration createField(Element element) {
		VariableDeclaration field = new ProposableVariableDeclarationImpl(element);
		resource.getContents().add(field);
		return field;
	}

	private VariableDeclaration createType(Type type) {
		VariableDeclaration global = new ProposableVariableDeclarationImpl(type);
		resource.getContents().add(global);
		return global;
	}
	
	private VariableDeclaration createVariableDeclaration(String name) {
		VariableDeclaration var = FramScriptFactory.eINSTANCE.createVariableDeclaration();
		var.setName(name);
		resource.getContents().add(var);
		return var;
	}
	
	private QualifiedExpression getParent(QualifiedExpression expr) {
		EStructuralFeature feature = expr.eContainingFeature();
		if (feature.equals(QUALIFIED_EXPRESSION__CHILD)) {
			return (QualifiedExpression) expr.eContainer();
		} else if (feature.equals(QUALIFIED_EXPRESSION__PARENT) 
				|| feature.equals(ARRAY_ELEMENT_EXPRESSION__ARRAY)) {
			return getParent((QualifiedExpression) expr.eContainer());
		} else {
			return null;
		}
	}
	
	private Block getBlock(EObject object) {
		while (object != null && !(object instanceof Block)) {
			object = object.eContainer();
		}
		return (Block) object;
	}
	
	private Function getFunction(EObject object) {
		while (object != null && !(object instanceof Function)) {
			object = object.eContainer();
		}
		return (Function) object;
	}
	
	private String getExtension(EObject object) {
		return object.eResource().getURI().fileExtension();
	}
	
}
